#  Copyright (c) 2023. #  OCX Consortium https://3docx.org. See the LICENSE
"""Python xml-rpc client for the OCX wiki."""

# System imports
from typing import Dict
import datetime
from pathlib import Path

# Third party imports
from dokuwiki import DokuWiki, DokuWikiError
from loguru import logger

# Module imports
import ocx_common.clients.wiki_struct_data as struct_data


class OcxWikiError(Exception):
    """Exception raised by this module when there is an error."""

    pass


class WikiClient:
    """The WikiClient provides functionality for interacting with the OCX wiki pages.

    Args:
        url: the OCX wiki URL
        user: The wiki user
        password: The user password

    Attributes:
        _url: the wiki url
        _wiki: the DocuWiki proxy

    """

    def __init__(self, url: str, user: str, password):
        self._url = url
        self._wiki = None
        self._connected = False
        try:
            self._wiki = DokuWiki(url, user, password)
        except DokuWikiError as e:
            logger.error(f"Connecting to {url} failed: {e}")
        except Exception as e:
            logger.error(f"Connecting to {url} failed: {e}")
        if self._wiki is None:
            raise DokuWikiError(f"Failed to connect to {url}")
        else:
            self._connected = True
            logger.info(f"Connected to {self._url}")
            logger.info(f"Dokuwiki version: {self._wiki.version}")
            logger.info(f"XMLRPC version: {self._wiki.xmlrpc_version}")

    def is_connected(self) -> bool:
        """True if a connection to the ocxwiki is established, False otherwise."""
        return self._connected

    def current_url(self) -> str:
        """Return the current ocxwiki url."""
        return self._url

    def login(self, user: str, password: str) -> bool:
        """Log in to the wiki.

        Args:
            user: the wiki user
            password: The user password
        """

        try:
            self._wiki.login(user, password)
            self._connected = True
            return True
        except (DokuWikiError, Exception) as err:
            logger.error(f"Unable to connect: {err}")
            self._connected = False
        return self._connected

    def wiki_version(self) -> str:
        """Return the ocxwiki dokuwiki version."""
        return self._wiki.version

    def xmlrpc_version(self) -> str:
        """Return the ocxwiki xmlrpc version."""
        return self._wiki.xmlrpc_version

    # Wiki pages
    def list_pages(
        self,
        namespace: str,
        depth: int = 0,
        md5_hash: bool = False,
        skip_acl: bool = False,
    ):
        """List all pages.

        Arguments:
            depth: recursion level, 0 for all
            md5_hash: do an md5 sum of content
            skip_acl: skip everything regardless of ACL
            namespace: List pages in this namespace.
        """
        options = {"depth": depth, "hash": md5_hash, "skipacl": skip_acl}
        return self._wiki.pages.list(namespace, **options)

    def changes(self, timestamp: datetime):
        """Returns a list of changes since given timestamp.

        Arguments:
            timestamp: input time

        Returns:
            Returns a list of changes since given timestamp.
        """
        return self._wiki.pages.changes(timestamp)

    def append_page(
        self, page: str, content: str, summary: str, namespace: str, minor: bool = False
    ):
        """Appends content to ''page''.

        Arguments:
            page: page name
            namespace: the namespace of the page
            content: content to be appended to the page
            summary: Change summary
            minor: Whether this is a minor change


        Returns:

        """
        wiki_page = f"{namespace}:{page}"
        result = False
        try:
            result = self._wiki.pages.append(wiki_page, content, summary, minor)
        except DokuWikiError as e:
            logger.error(e)
        return result

    def set_page(
        self, page: str, content: str, summary: str, namespace: str, minor: bool = False
    ) -> bool:
        """Set/replace the content of ''page''.

        Arguments:
            page: page name
            namespace: the namespace of the page
            content: content to be appended to the page
            summary: Change summary
            minor: Whether this is a minor change


        Returns:
            Returns True if the page was successfully set, False otherwise.
        """
        options = {"sum": summary, "minor": minor}
        wiki_page = f"{namespace}:{page}"
        result = False
        try:
            result = self._wiki.pages.set(wiki_page, content, **options)
        except DokuWikiError as e:
            logger.error(e)
        return result

    # Data structs
    def get_data(self, page: str, keep_order: bool = True) -> Dict:
        """Get the structured data of the latest version of a given page.

        Arguments:
            page: the page of interest
            keep_order: Return an ordered dict if True
        """
        # Get the page content
        data = {}
        try:
            content = self._wiki.pages.get(page)
            data = struct_data.struct_get(content, keep_order)
        except DokuWikiError as e:
            logger.error(e)
        return data

    # Wiki media
    def list_media(
        self,
        namespace: str,
        depth: int = 0,
        md5_hash: bool = False,
        skip_acl: bool = False,
        pattern: str = "*",
    ) -> Dict:
        """List the wiki media.

        Arguments:
            namespace: List media in namespace
            depth: recursion level, 0 for all
            md5_hash: do an md5 sum of content
            skip_acl: skip everything regardless of ACL
            pattern: list only media matching pattern
        """
        options = {"depth": depth, "hash": md5_hash, "skipacl": skip_acl}
        result = {}
        try:
            result = self._wiki.medias.list(namespace, **options)
        except DokuWikiError as e:
            logger.error(e)
        return result

    def media_changes(self, timestamp: datetime):
        """Returns the list of medias changed since given ''timestamp''.

        Arguments:
            timestamp: input time

        Returns:
            Returns a list of changes since given ''timestamp''.
        """
        return self._wiki.medias.changes(timestamp)

    def media_info(self, media: str):
        """Returns information of ''media''.

        Arguments:
            media: name of media

        Returns:
            Returns information of ''media''.
        """
        return self._wiki.medias.info(media)

    def add_media(self, media: str, filepath: Path, overwrite: bool = True):
        """Set media from local file filepath.

        Arguments:
            media: name of media
            filepath: path to local media file
            overwrite: parameter specify if the media must be replaced if it exists remotely.

        """

        self._wiki.medias.add(media, filepath.resolve(), overwrite)

    def media_delete(self, media: str):
        """Delete ''media''.

        Arguments:
            media: name of media

        """
        return self._wiki.medias.delete(media)



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE

"""Manage structured data on wiki pages."""

# System imports
from typing import Dict
import re
from dataclasses import dataclass, fields, field

# Third party imports
# Module imports
from ocx_common.clients.wiki_client import OcxWikiError


@dataclass
class BaseDataClass:
    """Base class for OCX dataclasses.

    Each subclass has to implement a field metadata with name `header` for each of its attributes, for example:

        ``name : str = field(metadata={'header': '<User friendly field name>'})``

    """

    def to_dict(self) -> Dict:
        """Output the data class as a dict with field names as keys."""
        my_fields = fields(self)
        return {
            my_fields[i].metadata["header"]: value
            for i, (key, value) in enumerate(self.__dict__.items())
        }


@staticmethod
def struct_get(content, keep_order=False) -> Dict:
    """Get dataentry from *content*. *keep_order* indicates whether to
    return an ordered dictionnay."""
    if keep_order:
        from collections import OrderedDict

        dataentry = OrderedDict()
    else:
        dataentry = {}

    found = False
    for line in content.split("\n"):
        if line.strip().startswith("---- dataentry"):
            found = True
            continue
        elif line == "----":
            break
        elif not found:
            continue

        line_split = line.split(":")
        key = line_split[0].strip()
        value = re.sub("#.*$", "", ":".join(line_split[1:])).strip()
        dataentry.setdefault(key, value)

    if not found:
        raise OcxWikiError("no dataentry found on the page")
    return dataentry


@staticmethod
def struct_gen(name: str, data: Dict) -> str:
    """Generate dataentry *name* from *data*."""
    struct = f"---- dataentry {name} ----\n"
    for attr, value in data.items():
        struct += f"\n{attr} : {value}"
    struct += "\n----\n"
    return struct


@staticmethod
def struct_ignore(content: str) -> str:
    """Remove dataentry from *content*."""
    page_content = []
    start = False
    for line in content.split("\n"):
        if line == "----" and not start:
            start = True
            continue
        if start:
            page_content.append(line)
    return "\n".join(page_content) if page_content else content


@dataclass
class WikiSchema(BaseDataClass):
    """Data class defining the structured data schema

    Parameters:
        ocx_version: OCX Schema version
        ocx_location: The uri of the schema location
        namespace: The schema type namespace
        author: Publishing author
        date: Publish date
        status: The schema status (draft or published)
        wiki_version: The wiki CLI version

    """

    ocx_version: str = field(metadata={"header": "OCX Version "})
    ocx_location: str = field(metadata={"header": "Location URL_url "})
    namespace: str = field(metadata={"header": "Namespace"})
    author: str = field(metadata={"header": "Author "})
    date: str = field(metadata={"header": "Date "})
    status: str = field(
        metadata={"header": "Status "}
    )  # For some strange reason the last character disappear when rendered on the wiki
    wiki_version: str = field(metadata={"header": "Publisher version "})



#  Copyright (c) 2024. #  OCX Consortium https://3docx.org. See the LICENSE
from pathlib import Path
# Project imports
from ocx_common.parser.parsers import OcxNotifyParser, OcxModelParser, XsdParser

from .conftest import MODEL_FOLDER, TEST_MODEL, NAMESPACE


class TestNotifyParser:
    def test_parse(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxNotifyParser()
        root = parser.parse(str(file))
        name = root.header.name
        assert name == "OCX-MODEL1/A"


class TestOcxParser:
    def test_parse(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxModelParser(str(file))
        root = parser.get_root()
        assert (
            root.tag
            == "{https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd}ocxXML"
        )


class TestXSDParser:
    def test_parse(self, shared_datadir):
        schema_path = shared_datadir / "schemas" / "OCX_Schema.xsd"
        schema_parser = XsdParser(target_namespace="https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd",
                                  location="https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd")
        schema_content = schema_path.read_text()
        schema = schema_parser.parse_from_string(schema_content)
        assert schema.location == "https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd"

# def test_parser_invalid_source(shared_datadir):   # ToDO: Add edge test case when source does not exist
#     file = shared_datadir / "not_exist.3docx"
#     try:
#         OcxModelParser(str(file))
#         assert False
#     except OcxParserError:
#         assert True



#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE


import json

# Project imports
from ocx_common.parser.parsers import OcxNotifyParser
from ocx_common.serializer.serializer import OcxSerializer
from tests.conftest import TEST_MODEL


def test_serialize_json(shared_datadir, data_regression):
    parser = OcxNotifyParser()
    model = shared_datadir / "models" / TEST_MODEL
    ocxxml = parser.parse(str(model.resolve()))
    serializer = OcxSerializer(ocxxml)
    result = serializer.serialize_json()
    data_regression.check(json.loads(result))


def test_serialize_xml(shared_datadir):
    parser = OcxNotifyParser()
    model = shared_datadir / "models"/ TEST_MODEL
    ocxxml = parser.parse(str(model.resolve()))
    serializer = OcxSerializer(ocxxml)
    result = serializer.serialize_xml()
    assert '?xml version' in result



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""This module provides the ocx_attribute_reader app functionality."""

# System imports
from pathlib import Path
from typing import Any, Tuple

# 3rd party imports
import typer
from typing_extensions import Annotated

from ocxtools import config
from ocxtools.context.context_manager import get_context_manager
from ocxtools.exceptions import XmlParserError
from ocxtools.parser.parser import OcxParser
from ocxtools.serializer import __app_name__

# Project imports
from ocxtools.serializer.serializer import OcxSerializer

JSON_INDENT = int(config.get("SerializerSettings", "json_indent"))
SERIALIZER_SUFFIX = config.get("SerializerSettings", "suffix")


serialize = typer.Typer(help="Serialisation of 3Docx models.")
ocx_parser = OcxParser()


@serialize.command()
def xml(
    model: Path,
    post_fix: Annotated[
        str, typer.Option(help="The model file name post fix")
    ] = SERIALIZER_SUFFIX,
):
    """Serialize a 3Docx model to file"""
    context_manager = get_context_manager()
    console = context_manager.get_console()
    try:
        ocx_obj = ocx_parser.parse(str(model))
        # Serialize
        output = model.parent.joinpath(f"{model.stem}{post_fix}{model.suffix}")
        output = output.resolve()
        serializer = OcxSerializer(ocx_obj)
        with output.open("w") as fp:
            fp.write(serializer.serialize_xml())
        console.print(f"Pretty printed file {output!r}")
        # Check for any warnings issued during serialization
        # for warning_message in warnings:
        #     print(f"Warning: {warning_message}")

    except XmlParserError as e:
        console.error(e)


@serialize.command()
def json(
    model: Path,
):
    """Serialize an 3Docx model to json."""
    context_manager = get_context_manager()
    console = context_manager.get_console()
    try:
        ocx_obj = ocx_parser.parse(str(model))
        # Serialize
        output = model.parent.joinpath(f"{model.stem}.json")
        output = output.resolve()
        serializer = OcxSerializer(ocx_obj)
        with output.open("w") as fp:
            fp.write(serializer.serialize_json())
        console.info(f"Serialised JSON to file {str(output)!r}")
    except XmlParserError as e:
        console.error(e)


def cli_plugin() -> Tuple[str, Any]:
    """
    ClI plugin

    Returns the typer command object
    """
    typer_click_object = typer.main.get_command(serialize)
    return __app_name__, typer_click_object



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""OcxSerializer module."""

# system imports
from dataclasses import dataclass
from enum import Enum


# 3rd party imports
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.serializers import JsonSerializer, XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig


# Project imports
from ocx_common.parser.parsers import MetaData


class ReportFormat(Enum):
    """Serialisation formats"""

    CSV = "csv"
    PARQUET = "parquet"


class OcxSerializer:
    """OcxSerializer class for 3Docx XML models."""

    def __init__(
        self,
        ocx_model: dataclass,
        pretty_print: bool = True,
        pretty_print_indent: str = "  ",
        encoding: str = "utf-8",
    ):
        """
        Args:
            ocx_model: The dataclass to serialize.
            pretty_print: True to pretty print, False otherwise.
            pretty_print_indent: Pretty print indentation.
            encoding: The encoding code.
        Params:
            _model: The dataclass to serialize.
            _config: The serializer configuration.

        """
        self._model: dataclass = ocx_model
        self._config = SerializerConfig(
            encoding=encoding,
            xml_version="1.0",
            xml_declaration=True,
            pretty_print=True,
            ignore_default_attributes=False,
            schema_location=None,
            no_namespace_schema_location=None,
            globalns=None,
        )

    def serialize_xml(self, global_ns: str = "ocx") -> str:
        """Serialize a 3Docx XML file with proper indentations.

        Returns:
              The dataclass xml serialisation.

        Raises:
            SerializeError if failing
        """
        target_ns = MetaData.namespace(self._model)
        ns_map = {global_ns: target_ns}
        serializer = XmlSerializer(context=XmlContext(), config=self._config)
        return serializer.render(self._model, ns_map=ns_map)

    def serialize_json(self) -> str:
        """Serialize a 3Docx XML model to json with proper indentations.

        Returns:
              The dataclass xml serialisation.

        Raises:
            SerializeError if failing
        """
        serializer = JsonSerializer(context=XmlContext(), config=self._config)
        return serializer.render(self._model)


# class Serializer:
#     """A general serializer for dict type data structures"""
#
#     @staticmethod
#     def serialize_to_csv(table: List, file_name: str):
#         """
#         Serialize a list of dictionaries to a csv file. Each dictionary is a row with ``key:value`` pairs
#         where the key is the column header and the value is the data value.
#         Args:
#             table: The table to serialize
#             file_name: the output file name
#         """
#         with open(file_name, 'w', newline='') as csvfile:
#             fieldnames = table[0].keys()
#             writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
#             writer.writeheader()
#             writer.writerows(table)
#
#     @staticmethod
#     def serialize_to_parquet(report: ReportDataFrame, report_folder: str):
#         """
#         Serialize a dataframe report to a parquet file
#         Args:
#             report: Dataclass containing the dataframe to serialize
#             report_folder: the output directory
#         """
#         try:
#             stem = Path(report.source).stem
#             ocx_type = report.type.value.lower()
#             file_name = Path(report_folder).joinpath(f'{stem}_{ocx_type}.parquet')
#             df = report.elements
#             df.to_parquet(file_name)
#         except (pyarrow.lib.ArrowException, pyarrow.lib.ArrowInvalid) as e:
#             logger.error(f'Error serializing to parquet: {e}')
#             raise SerializerError(e) from e
#


class SerializerError(ValueError):
    """OCX Serializing errors."""
